///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      NTSRVICE.CPP
//
//  Description
//      This file implements functions specific for NT services stuff like
//      install, remove, ServiceMain, etc.
//      It also implement the function to help install and remove
//      configuration information.
//
//  Author
//      Irving De la Cruz (irvingd@microsoft.com)
//
//  Revision: 1.1
//
// Written for Microsoft Corporation.
// Copyright (c) 1996-1997 Microsoft Corporation. All rights reserved.
//
#include "DIRMON.H"
#include "NTSECAPI.H"
#include <LM.H>

///////////////////////////////////////////////////////////////////////////////
//    InstallService()
//
//    Parameters
//      [IN]    szFullPathToServiceExe  Full path with EXE name of the service
//      [IN]    szServiceName           Display name of the service. This is
//                                      the name displayed in the SCM control
//                                      panel applet.
//      [IN]    szServiceNameShort      Short name of the service. This is not
//                                      displayed in the SCM control panel
//                                      applet.
//      [IN]    szAccount               Account under which the service runs.
//                                      This account in the form "DOMAIN\USER"
//                                      if the account is local, then it must
//                                      be in the form ".\USER".
//      [IN]    szPassword              Password for the szAccount
//
//    Purpose
//      Install a Windows NT service with the Service Control Manager (SCM)
//
//    Return Value
//      A Win32 error code
//
DWORD WINAPI InstallService (LPSTR szFullPathToServiceExe,
                             LPSTR szServiceName,
                             LPSTR szServiceNameShort,
                             LPSTR szAccount,
                             LPSTR szPassword)
{
    DWORD dwError = NO_ERROR;
    SC_HANDLE schSCM, schService;
    schSCM = OpenSCManager (0, 0, SC_MANAGER_CREATE_SERVICE);
    if (NULL == schSCM)
    {
        dwError = GetLastError();
    }
    else
    {
RestartInstall:
        schService = CreateService (schSCM,
                                    szServiceNameShort,
                                    szServiceName,
                                    SERVICE_ALL_ACCESS,
                                    SERVICE_WIN32_OWN_PROCESS |
                                    0,
                                    SERVICE_DEMAND_START,
                                    SERVICE_ERROR_NORMAL,
                                    szFullPathToServiceExe,
                                    NULL,
                                    NULL,
                                    NULL,
                                    szAccount,
                                    szPassword);
        if (NULL == schService)
        {
            dwError = GetLastError();
        }
        else
        {
            CloseServiceHandle (schService);
        }
        if (ERROR_SERVICE_EXISTS == dwError)
        {
            RemoveService (szServiceNameShort);
            dwError = NO_ERROR;
            goto RestartInstall;
        }
        CloseServiceHandle (schSCM);
    }
    return dwError;
}

///////////////////////////////////////////////////////////////////////////////
//    RemoveService()
//
//    Parameters
//      [IN]    szServiceNameShort  The short name of the service
//
//    Purpose
//      Removes the specified service from the service control manager.
//      
//    Return Value
//      A Win32 Error code
//
DWORD WINAPI RemoveService (LPSTR szServiceNameShort)
{
    DWORD dwError = NO_ERROR;
    SC_HANDLE schSCM, schService;
    schSCM = OpenSCManager (0, 0, SC_MANAGER_CREATE_SERVICE);
    if (NULL == schSCM)
    {
        dwError = GetLastError();
    }
    else
    {
        schService = OpenService (schSCM,
                                  szServiceNameShort,
                                  SERVICE_ALL_ACCESS);
        if (NULL == schService)
        {
            dwError = GetLastError();
        }
        else
        {
            SERVICE_STATUS ssState;
            ControlService (schService, SERVICE_CONTROL_STOP, &ssState);
            DeleteService (schService);
            CloseServiceHandle (schService);
        }
        CloseServiceHandle (schSCM);
    }
    return dwError;
}

///////////////////////////////////////////////////////////////////////////////
//    AddConfigRegistryValues()
//
//    Parameters
//      [IN]    pszMAPIProfile      The MAPI profile for the service to use
//      [IN]    pszMonitoredDir     The directory the service will monitor
//
//    Purpose
//      Adds the configuration parameters of the service to the registry.
//
//    Return Value
//      A Win32 error code
//      
DWORD WINAPI AddConfigRegistryValues (LPSTR pszMAPIProfile,
                                      LPSTR pszMonitoredDir)
{
    HKEY hKey;
    DWORD dwDisposition;
    DWORD dwError = RegCreateKeyEx (HKEY_LOCAL_MACHINE,
                                    REGKEY_DIRMON_SERVICE_CONFIG,
                                    0,
                                    NULL,
                                    REG_OPTION_NON_VOLATILE,
                                    KEY_WRITE,
                                    NULL,
                                    &hKey,
                                    &dwDisposition);
    if (ERROR_SUCCESS != dwError)
    {
        return dwError;
    }
    RegSetValueEx (hKey,
                   REGKEY_CONFIG_PROFILE,
                   0,
                   REG_SZ,
                   (LPBYTE)pszMAPIProfile,
                   lstrlen (pszMAPIProfile));
    RegSetValueEx (hKey,
                   REGKEY_CONFIG_MONDIR,
                   0,
                   REG_SZ,
                   (LPBYTE)pszMonitoredDir,
                   lstrlen (pszMonitoredDir));
    RegCloseKey (hKey);
    return dwError;
}

///////////////////////////////////////////////////////////////////////////////
//    AddEventLogSource()
//
//    Parameters
//      [IN]    szServiceEXE    Full path to the service EXE
//
//    Purpose
//      Add an event log source where the messages are located. The message
//      compiler string table is part of the resource table of the service
//      executable.
//      
//    Return Value
//      A Win32 error code
//      
DWORD WINAPI AddEventLogSource (LPSTR szServiceEXE)
{
    HKEY hKey;
    DWORD dwDisposition;
    DWORD dwError = RegCreateKeyEx (HKEY_LOCAL_MACHINE,
                                    "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\"DIRMON_EVENTLOG_SOURCE,
                                    0,
                                    NULL,
                                    REG_OPTION_NON_VOLATILE,
                                    KEY_WRITE,
                                    NULL,
                                    &hKey,
                                    &dwDisposition);
    if (ERROR_SUCCESS != dwError)
    {
        return dwError;
    }

    DWORD cbSize, dwData;
    cbSize = sizeof(DWORD);
    dwData = 1;
    RegSetValueEx (hKey,
                   "CategoryCount",
                   0,
                   REG_DWORD,
                   (LPBYTE)&dwData,
                   cbSize);
    dwData = 0x1F;
    RegSetValueEx (hKey,
                   "TypesSupported",
                   0,
                   REG_DWORD,
                   (LPBYTE)&dwData,
                   cbSize);
    RegSetValueEx (hKey,
                   "CategoryMessageFile",
                   0,
                   REG_EXPAND_SZ,
                   (LPBYTE)szServiceEXE,
                   lstrlen(szServiceEXE) + 1);
    RegSetValueEx (hKey,
                   "EventMessageFile",
                   0,
                   REG_EXPAND_SZ,
                   (LPBYTE)szServiceEXE,
                   lstrlen(szServiceEXE) + 1);
    RegCloseKey (hKey);
    return dwError;
}

///////////////////////////////////////////////////////////////////////////////
//    RemoveEventLogSource()
//
//    Parameters
//      None
//
//    Purpose
//      Remove the event log source for this service
//      
//    Return Value
//      A Win32 error code
//      
DWORD WINAPI RemoveEventLogSource()
{
    HKEY hKey;
    DWORD dwError = RegOpenKeyEx (HKEY_LOCAL_MACHINE,
                                  "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application",
                                  0,
                                  KEY_ALL_ACCESS,
                                  &hKey);
    if (ERROR_SUCCESS == dwError)
    {
        RegDeleteKey (hKey, DIRMON_EVENTLOG_SOURCE);
        RegCloseKey (hKey);
    }
    return dwError;
}

///////////////////////////////////////////////////////////////////////////////
//    DirMonServiceMain()
//
//    Parameters
//      { Refer to Win32 API documentation on ServiceMain }
//
//    Purpose
//      Entry point called by the Windows NT service control manager to
//      start the service and obtain the command handler function callback.
//      With the callback the SCM will tell the service the next state it
//      should be in (i.e. pause, running, stopped, etc.)
//      
//    Return Value
//      None
//      
void DirMonServiceMain (DWORD dwArgc, LPTSTR * pszArgv)
{
    InitEventLog();
    LogEvent (FALSE, EVENTMSG_I_SERVICE_STARTING);
    
    g_ssHandle = RegisterServiceCtrlHandler (DIRMON_SERVICE_NAME_SHORT,
                                             (LPHANDLER_FUNCTION)DirMonServiceCtrlHandler);
    if (NULL == g_ssHandle)
    {
        LogEvent (TRUE, EVENTMSG_E_FAILED_REGISTRER_CTRL_HANDLER, GetLastError());
        return;
    }
    DWORD dwError = SendStatusToSCM (SERVICE_START_PENDING, S_OK, 1, 5);
    if (dwError)
    {
        LogEvent (TRUE, EVENTMSG_E_FAILED_NOTIFY_SCM_STARTING, dwError);
        TerminateService (dwError);
        return;
    }
    g_hTerminateEvent = CreateEvent (0, TRUE, FALSE, 0);
    if (NULL == g_hTerminateEvent)
    {
        dwError = GetLastError();
        LogEvent (TRUE, EVENTMSG_E_FAILED_CREATE_EVENT, dwError);
        TerminateService (dwError);
        return;
    }

    dwError = SendStatusToSCM (SERVICE_START_PENDING, S_OK, 2, 15);
    if (!dwError)
    {
        dwError = InitializeService (NULL);
        if (!dwError)
        {
            SendStatusToSCM (SERVICE_RUNNING, S_OK, 0, 0);
        }
    }
    if (dwError)
    {
        LogEvent (FALSE, EVENTMSG_E_FAILED_TO_START_SERVICE, dwError);
    }
    else
    {
        LogEvent (FALSE, EVENTMSG_I_SERVICE_STARTED);
        ServiceWaitOnEvent (g_hTerminateEvent);
        FinishDirectoryWatch();
        LogEvent (FALSE, EVENTMSG_I_SERVICE_STOPPED);
    }
    TerminateService (dwError);
}

///////////////////////////////////////////////////////////////////////////////
//    ServiceWaitOnEvent()
//
//    Parameters
//      [IN]    hEvent  Event we must wait on
//
//    Purpose
//      Waits on the specified event until it is set. Signaling the event
//      means the service is terminating thus, after the wait is over, the
//      function returns.
//      
//    Return Value
//      A Win32 Error code.
//      
DWORD WINAPI ServiceWaitOnEvent (HANDLE hEvent)
{
    // We must not block with WaitForSingleObject() in this thread because this thread
    // has initialized OLE. OLE creates hidden windows which need to process globally
    // sent messages (i.e. DDE broadcasts) otherwise the application that sent it
    // will hang.
    DWORD dwWait;
    MSG msg;
    BOOL fContinue = TRUE;
    do
    {
        dwWait = MsgWaitForMultipleObjects (1, &hEvent, FALSE, INFINITE, QS_ALLINPUT);
        if (WAIT_OBJECT_0 + 1 == dwWait)
        {
            // Get the message out of the queue
            while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
            {
                TranslateMessage (&msg);
                DispatchMessage (&msg);
            }
        }
        else
        {
            fContinue = FALSE;
        }
    } while (fContinue);
    return NO_ERROR;
}

///////////////////////////////////////////////////////////////////////////////
//    SendStatusToSCM()
//
//    Parameters
//      [IN]    dwServiceState      New state of the service
//      [IN]    dwError             Error code for the current state
//      [IN]    dwCheckPoint        Current progress, a fraction of dwWait
//      [IN]    dwWait              How long the process will take
//
//    Purpose
//      This function interacts with the service control manager to inform
//      it of new state and error code that occur in the service.
//      
//    Return Value
//      A Win32 Error code.
//      
DWORD WINAPI SendStatusToSCM (DWORD dwServiceState,
                              DWORD dwError,
                              DWORD dwCheckPoint,
                              DWORD dwWait)
{
    SERVICE_STATUS ssStatus = { 0 };
    ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    ssStatus.dwCurrentState = dwServiceState;
    if (SERVICE_START_PENDING == dwServiceState)
    {
        ssStatus.dwControlsAccepted = 0;
    }
    else
    {
        ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP |
                                      SERVICE_ACCEPT_SHUTDOWN;
    }
    if (dwError)
    {
        ssStatus.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
        ssStatus.dwServiceSpecificExitCode = dwError;
    }
    ssStatus.dwCheckPoint = dwCheckPoint;
    ssStatus.dwWaitHint = (dwWait * 1000);
    
    HRESULT dwInteralError = 0;
    if (!SetServiceStatus (g_ssHandle, &ssStatus))
    {
        dwInteralError = GetLastError();
    }
    return dwInteralError;
}

///////////////////////////////////////////////////////////////////////////////
//    DirMonServiceCtrlHandler()
//
//    Parameters
//      [IN]    dwControlCode   Control code give to us by the SCM
//
//    Purpose
//      Given a control code, we change the service state and inform the
//      service controller of our success and new state.
//      
//    Return Value
//      None
//      
void DirMonServiceCtrlHandler (DWORD dwControlCode)
{
    switch (dwControlCode)
    {
        case SERVICE_CONTROL_SHUTDOWN :
        case SERVICE_CONTROL_STOP :
            SendStatusToSCM (SERVICE_STOPPED, S_OK, 0, 0);
            SetEvent (g_hTerminateEvent);
            break;
        case SERVICE_CONTROL_INTERROGATE :
            SendStatusToSCM (0, S_OK, 0, 0);
            break;
    }
}

///////////////////////////////////////////////////////////////////////////////
//    InitLsaString()
//
//    Parameters
//      [IN]    pucstrString    A caller-allocated structure for
//                              the priviledge-granting functions
//      [IN]    pszString       The string used to initialize
//                              the LSA structure
//      
//    Purpose
//      Initializes the LSA structure with the UNICODE string
//      passed in pszString.
//      
//    Return Value
//      None
//
void WINAPI InitLsaString (PLSA_UNICODE_STRING pucstrString, LPWSTR pszString)
{
	DWORD dwLen = lstrlenW (pszString);
	pucstrString->Buffer = pszString;
	pucstrString->Length = (USHORT)(dwLen * sizeof(WCHAR));
	pucstrString->MaximumLength = (USHORT)((dwLen + 1) * sizeof(WCHAR));
}

///////////////////////////////////////////////////////////////////////////////
//    SetRequiredPrivileges()
//
//    Parameters
//      [IN]    pszUserAccount	    NT user account name without the DOMAIN
//      
//    Purpose
//      Add the neccessary priviledges to the specified account in
//      pszUserAccount to log in to a MAPI profile from the context of an
//      NT service. The function will grant the following priviledges:
//
//          Restore Files and Directories (SE_RESTORE_NAME)
//          Logon as a Service (SE_SERVICE_LOGON_NAME)
//      
//    Return Value
//      A Win32 error code
//
DWORD WINAPI SetRequiredPrivileges (LPSTR pszUserAccount)
{
	LSA_HANDLE hPolicy;
	LSA_OBJECT_ATTRIBUTES lsaObjAttributes;
	LSA_UNICODE_STRING ucstrServer, ucstrPrivileges[2];
	PSID pSid = NULL;
	SID_NAME_USE sidUse;
	NTSTATUS status;
	WCHAR szPrivilege1[80], szPrivilege2[80], wcsCompName[MAX_COMPUTERNAME_LENGTH+1];
	char szDomain[80];
	DWORD dwResult, cbSid = 0, cbDomain = 79;

	// Retrieve SID and domain associated w/NT user account
	LookupAccountName(NULL, pszUserAccount, NULL, &cbSid, szDomain, &cbDomain, &sidUse);
	if ((dwResult = GetLastError()) != ERROR_INSUFFICIENT_BUFFER)
    {
		return dwResult;
    }
	pSid = (PSID) new char[cbSid];
	if (!LookupAccountName(NULL, pszUserAccount, pSid, &cbSid, szDomain, &cbDomain, &sidUse))
    {
		return GetLastError();
    }

    char szThisCompName[MAX_COMPUTERNAME_LENGTH+1];
    DWORD dwNameSize = MAX_COMPUTERNAME_LENGTH+1;
    GetComputerName (szThisCompName, &dwNameSize);

	// Retrieve the PDC associated w/user account
	mbstowcs(wcsCompName, szThisCompName, MAX_COMPUTERNAME_LENGTH+1);
	// Open policy database
	ZeroMemory (&lsaObjAttributes, sizeof(lsaObjAttributes));
	InitLsaString (&ucstrServer, wcsCompName);
	if ((status = LsaOpenPolicy (&ucstrServer, 
								 &lsaObjAttributes, 
								 POLICY_CREATE_ACCOUNT | POLICY_LOOKUP_NAMES,
								 &hPolicy)) == NO_ERROR)
	{
		// Grant privileges to user account
	    mbstowcs (szPrivilege1, SE_RESTORE_NAME, 79);
        mbstowcs (szPrivilege2, SE_SERVICE_LOGON_NAME, 79);
	    InitLsaString (&ucstrPrivileges[0], szPrivilege1);
	    InitLsaString (&ucstrPrivileges[1], szPrivilege2);
	    status = LsaAddAccountRights (hPolicy, pSid, ucstrPrivileges, 2);
	    LsaClose(hPolicy);
    }
	delete [] pSid;
	return (DWORD)status;
}


///////////////////////////////////////////////////////////////////////////////
//    AddUserToLocalGroup()
//
//    Parameters
//      [IN]    pszAccountName	    NT account name (e.g., irvingd)
//		[IN]    pszLocalGroup	    Name of local group (e.g., Administrators)
//      
//    Purpose
//      Add domain NT account name to a local group
//      
//    Return Value
//      A Win32 Error code
//
DWORD WINAPI AddUserToLocalGroup (LPSTR pszAccountName, LPSTR pszLocalGroup)
{
    DWORD dwError = NO_ERROR;
	DWORD cbSid = 0, cbDomain = 127;
	TCHAR szDomain[128];
	WCHAR szLocalGroup[256];
	SID_NAME_USE sidUse;
    PSID pSid = NULL;
	LookupAccountName (NULL,
                       pszAccountName,
                       pSid,
                       &cbSid,
                       szDomain,
                       &cbDomain,
                       &sidUse);
    pSid = (PSID)new char[cbSid];
    // Convert user account name to SID
    BOOL bResult = LookupAccountName (NULL,
                                      pszAccountName,
                                      pSid,
                                      &cbSid,
                                      szDomain,
                                      &cbDomain,
                                      &sidUse);
	if (bResult)
    {
	    // Convert ASCII local group name to UNICODE
	    mbstowcs (szLocalGroup, pszLocalGroup, 255);
        // Add user to local group
	    if (NetLocalGroupAddMember (NULL, szLocalGroup, pSid) != 0)
        {
            dwError = ERROR_MEMBER_IN_GROUP;
        }
    }
    else
    {
        dwError = GetLastError();
    }
	delete [] pSid;
    return dwError;
}

// End of file for NTSRVICE.CPP
